/********************************************************************
 *                                                                  *
 * THIS FILE IS PART OF THE QuickTime Components SOURCE CODE.       *
 * USE, DISTRIBUTION AND REPRODUCTION OF THIS LIBRARY SOURCE IS     *
 * GOVERNED BY A BSD-STYLE SOURCE LICENSE INCLUDED WITH THIS SOURCE *
 * IN 'COPYING'. PLEASE READ THESE TERMS BEFORE DISTRIBUTING.       *
 *                                                                  *
 * THE QuickTime Components SOURCE CODE IS (C) COPYRIGHT 2001-2002  *
 * by Steve Nicolai http://qtcomponents.sourceforge.net/            *
 *                                                                  *
 ********************************************************************

 function: QuickTime resource Ids for all components
 last mod: $$

 ********************************************************************/

// common header file shared by both the code and resource compilers

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// resource IDs

#ifndef __LAMECODEC_H__
#define __LAMECODEC_H__ 1

#ifdef NEVER_DEFINED

#define kDecompressorResID              1000    // ID of decompressor 'thng', code resources
#define kDecompressorNameStringResID    1000    // ID of codec name
#define kDecompressorInfoStringResID    1001    // ID of codec info string

#endif

#define kSoundCompressorResID           2000    // only used for dialogs, there is no component


#define kExporterResID                  3000
#define kExporterNameStringResID        3000
#define kExporterInfoStringResID        3001

#ifdef NEVER_DEFINED

#define kImporterResID                  4000
#define kImporterNameStringResID        4000
#define kImporterInfoStringResID        4001

#endif

#define kSoundComponentManufacturer     'PYEh'	// your company's OSType
#define kCodecFormat                    'mp3 '
//define kVorbisCodecFormat              'mp3 '

#define kLAMECodecHexVersion            0x00010002  // initial version
#define kLAMECodecVersion               "1.02"
#define kLAMECodecMacVERS               1, 0, development, 2
#define kLAMECodecWin32Version          1,0,0,0
#define kLAMECodecWin32ValVersion       "1,0,0,0\0"

#define kDecompressorVersion            0x00010000
#define kExporterVersion                0x00010002
#define kImporterVersion                0x00010002



// sound compression atom types (really exporter settings right now)
#define kLAMESettingsAtom               'Lame'
#define kLAMESettingsVersionAtom        'VERS'
#define kLAMESettingsQualityAtom		'QUAL'

#define kLAMESettings_vbrMode			'VBrM'
#define kLAMESettings_vbrQuality		'VBrQ'
#define kLAMESettings_stereoMode		'MODE'
#define kLAMESettings_sampleRate		'SRat'
#define kLAMESettings_bitRate			'RATE'
#define kLAMESettings_quality			'QalT'

#define kVorbisSettingsMinBitRateAtom     'MIN '
#define kVorbisSettingsAvgBitRateAtom     'AVG '
#define kVorbisSettingsMaxBitRateAtom     'MAX '
//@@@ may have to add additional settings for channel coupling


// resource names will be created for debug versions
#ifdef _DEBUG
#define LAMEDecompressorName    "LAME MP3 Decompressor"
#define LAMEExporterName        "LAME MP3 Exporter"
#define LAMEImporterName        "LAME MP3 Importer"
#else
#define LAMEDecompressorName    ""
#define LAMEExporterName        ""
#define LAMEImporterName        ""
#endif


#ifdef NEVER_DEFINED

#ifndef REZ
typedef struct Settings {
    long        version;                /* from settings */
    long        quality;                /* from settings */
    long        minBitRate;             /* from settings */
    long        desiredBitRate;	        /* from settings */
    long        maxBitRate;             /* from settings */
} Settings;
typedef Settings *SettingsPtr;
#endif

#endif

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// failure handling macros
/*
	Some macros used to check for errors and also to allow for
	handling them by using a goto statement.  This makes the source
	code easier to read.  It will break into the debugger with a
	message showing the condition that caused the failure.  In some
	of the macros the debug message is removed but goto remains.  In
	other macros all of it is removed when doing a non-debug build.

    Note that these macros use the "\p" construct for creating
	Pascal strings at compile time.  Most non-Mac compilers do
	not recognize this, give a warning, and put 'p' as the first
	character of the string.  You can ignore the warning because
	the non-Mac version of DebugStr deals with this just fine.
	For Microsoft's Visual C++, we suppress the warning below.
*/
#if defined(_MSC_VER) && !defined(__MWERKS__) 
	// Visual C++ from Microsoft
	#pragma warning(disable:4129) // unrecognized character escape sequence
#endif

// This checks for the exception, and if true then goto handler

#ifdef _DEBUG
#define FailIf(cond, handler)								\
	if (cond) {												\
		DebugStr((ConstStr255Param)"\p"#cond " goto " #handler);	\
		goto handler;										\
	} else 0
#else
#define FailIf(cond, handler)								\
	if (cond) {												\
		goto handler;										\
	} else 0
#endif

// This checks for the exception, and if true do the action and goto handler

#ifdef _DEBUG
#define FailWithAction(cond, action, handler)				\
	if (cond) {												\
		DebugStr((ConstStr255Param)"\p"#cond " goto " #handler);	\
		{ action; }											\
		goto handler;										\
	} else 0
#else
#define FailWithAction(cond, action, handler)				\
	if (cond) {												\
		{ action; }											\
		goto handler;										\
	} else 0
#endif

// This will insert debugging code in the application to check conditions
// and displays the condition in the debugger if true.  This code is
// completely removed in non-debug builds.

#ifdef _DEBUG
#define FailMessage(cond)									\
	if (cond)												\
		DebugStr((ConstStr255Param)"\p"#cond);				\
	else 0
#else
#define FailMessage(cond)
#endif

// This allows you to test for the result of a condition (i.e. CloseComponent)
// and break if it returns a non zero result, otherwise it ignores the result.
// When a non-debug build is done, the result is ignored.

#ifdef _DEBUG
#define ErrorMessage(cond)									\
	if (cond)												\
		DebugStr((ConstStr255Param)"\p"#cond);				\
	else 0
#else
#define ErrorMessage(cond)		cond
#endif

// This will display a given message in the debugger, this code is completely
// removed in non-debug builds.

#ifdef _DEBUG
#define DebugMessage(s)			DebugString((ConstStr255Param)s)
#else
#define DebugMessage(s)
#endif

#endif
