/********************************************************************
 *                                                                  *
 * THIS FILE IS PART OF THE QuickTime Components SOURCE CODE.       *
 * USE, DISTRIBUTION AND REPRODUCTION OF THIS LIBRARY SOURCE IS     *
 * GOVERNED BY A BSD-STYLE SOURCE LICENSE INCLUDED WITH THIS SOURCE *
 * IN 'COPYING'. PLEASE READ THESE TERMS BEFORE DISTRIBUTING.       *
 *                                                                  *
 * THE QuickTime Components SOURCE CODE IS (C) COPYRIGHT 2001-2002  *
 * by Steve Nicolai http://qtcomponents.sourceforge.net/            *
 *                                                                  *
 ********************************************************************

 function: QuickTime Movie Export component
 last mod: $$

 ********************************************************************/

#if defined(_MSC_VER) && !defined(__MWERKS__) 
#pragma warning(disable:4229)		// ignore anachronism used: modifiers on data are ignored
#endif

#if __APPLE_CC__
    #include <QuickTime/QuickTime.h>
#else
	#if defined(_WIN32)
	#include <windows.h>
	#endif

    #include <MacTypes.h>
    #include <MacErrors.h>
    #include <Endian.h>
    #include <MacMemory.h>
    #include <Resources.h>
    #include <Components.h>
    #include <Sound.h>
    #include <QuickTimeComponents.h>
    #include <FixMath.h>
    #include <NumberFormatting.h>
    #include <Dialogs.h>

    #if TARGET_OS_MAC
    #include <TextCommon.h>
    #include <TextEncodingConverter.h>
    #endif
    
    #if !TARGET_OS_MAC
    #include "QTML.h"
    #endif
#endif

#include <string.h>

#include "LAMECodec.h"
#include <LAME/lame.h>


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//			Constants
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

const int kConvertBufferSize = 4096;
const int kNumSamplesToFetch = 44100;

enum LAMEControlValues {
	kNone, 

	kOKButton, 
	kCancelButton, 

	kControls_START, 

	k_vbrMode	= kControls_START, 
	k_vbrQuality, 
    k_stereoMode,
    k_sampleRate,
	k_bitRate, 
	k_quality, 

	kControls_NUMTYPES
};

enum SoundConverterState {
	kReadingData,
	kSawEndofData,
	kEndedConversion
};

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//			types
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

typedef struct {
	vbr_mode	vbrMode;
	long		vbrQuality;
    MPEG_mode	stereoMode;
    long		sampleRate;
	long		bitRate;
	long		quality;
} LAME_Settings;

// Keep your data long word aligned for best performance

typedef struct {
    OSType                      trackType;
    long                        trackID;
    TimeScale                   scale;
    MovieExportGetPropertyUPP	getPropertyProc;
    MovieExportGetDataUPP       getDataProc;
    void                        *refcon;
    
    // the following are filled out from the data source property proc
    OSType                      compressionType;
    short                       channelCount;
    short                       sampleSize;
    UnsignedFixed               sampleRate;
} SoundDataSource;


typedef struct {
    ComponentInstance	self;
    
    /* the following are used to convert the sound to native endian 16-bit samples */
    SoundComponentData              theInputFormat;
    SoundComponentData              theOutputFormat;
    SoundConverter                  mySoundConverter;
    Ptr                             convertOutBuffer;
    SoundConverterFillBufferDataUPP fillBufferDataUPP;
    ExtendedSoundComponentData      sourceData;
    enum SoundConverterState        state;
    
    /* The following are used to fetch data from the source movie */
    MovieExportGetDataParams    getData;
    Boolean                     didFirstFetch;
    
    int                 numDataSources;
    SoundDataSource     source;
    
    MovieProgressUPP    progressProc;
    long                progressRefcon;
    
    Movie               sourceMovie;
    TimeValue           startTime;
    TimeValue           endTime;
    TimeValue           progressStart;
    TimeValue           progressDuration;
    int                 progressShift;
    
    OSErr               sourceError;
    
    Boolean             canceled;
    
    lame_global_flags   *lameflags;
    
    LAME_Settings       settings;
    
/*
    
    ogg_stream_state    os; // take physical pages, weld into a logical stream of packets //
    
    vorbis_info         vi; // struct that stores all the static vorbis bitstream settings //
    vorbis_comment      vc; // struct that stores all the user comments //
    
    vorbis_dsp_state    vd; // central working state for the packet->PCM decoder //
    vorbis_block        vb; // local working space for packet->PCM decode //
    */
    
    ComponentInstance   dataHandler;	/* output data handler */
    unsigned long       dataOffset;
    
    Handle              tagMappings;    /* TagM resource */
	
} LAMEExportGlobals, *LAMEExportGlobalsPtr;


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//			prototypes
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

/* sometimes I wonder if using ComponentDispatch.h is worth it.  This is
   one of those times.  MacOS functions are declared pascal, win32 are not.
 */
#if TARGET_OS_MAC
#define COMPONENTFUNC static pascal ComponentResult
#else
#define COMPONENTFUNC static ComponentResult
#endif

COMPONENTFUNC LAMEExportOpen(LAMEExportGlobalsPtr globals, ComponentInstance self);
COMPONENTFUNC LAMEExportClose(LAMEExportGlobalsPtr globals, ComponentInstance self);
COMPONENTFUNC LAMEExportVersion(LAMEExportGlobalsPtr globals);
COMPONENTFUNC LAMEExportToFile(LAMEExportGlobalsPtr globals, const FSSpec *theFile, 
                                            Movie m, Track onlyThisTrack, TimeValue startTime, TimeValue duration);
COMPONENTFUNC LAMEExportGetSourceMediaType (LAMEExportGlobalsPtr globals, OSType * mediaType);
COMPONENTFUNC LAMEExportValidate(LAMEExportGlobalsPtr globals, Movie theMovie, Track onlyThisTrack,
											Boolean *valid);
COMPONENTFUNC LAMEExportGetFileNameExtension(LAMEExportGlobalsPtr globals, OSType *extension);
COMPONENTFUNC LAMEExportGetShortFileTypeString(LAMEExportGlobalsPtr globals, Str255 typeString);
COMPONENTFUNC LAMEExportSetProgressProc(LAMEExportGlobalsPtr globals, MovieProgressUPP proc, long refcon);
COMPONENTFUNC LAMEExportAddDataSource (LAMEExportGlobalsPtr globals, OSType trackType, TimeScale scale,
		long *trackID, MovieExportGetPropertyUPP getPropertyProc, MovieExportGetDataUPP getDataProc, void *refCon);
COMPONENTFUNC LAMEExportFromProceduresToDataRef(LAMEExportGlobalsPtr globals, Handle dataRef, OSType dataRefType);
COMPONENTFUNC LAMEExportToDataRef(LAMEExportGlobalsPtr globals, Handle dataRef, OSType dataRefType, Movie theMovie, 
				Track onlyThisTrack,TimeValue startTime, TimeValue duration);
COMPONENTFUNC LAMEExportDoUserDialog(LAMEExportGlobalsPtr globals, Movie theMovie, Track onlyThisTrack,
									TimeValue startTime, TimeValue duration, Boolean *canceled);
COMPONENTFUNC LAMEExportGetSettingsAsAtomContainer(LAMEExportGlobalsPtr globals, QTAtomContainer *settings);
COMPONENTFUNC LAMEExportSetSettingsFromAtomContainer(LAMEExportGlobalsPtr globals, QTAtomContainer settings);


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//			Component Dispatcher
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#if TARGET_CPU_68K
	#define COMPONENT_C_DISPATCHER
	#define COMPONENT_DISPATCH_MAIN
#endif

#define CALLCOMPONENT_BASENAME() 		LAMEExport
#define CALLCOMPONENT_GLOBALS() 		LAMEExportGlobalsPtr storage

#define MOVIEEXPORT_BASENAME() 			CALLCOMPONENT_BASENAME()
#define MOVIEEXPORT_GLOBALS() 			CALLCOMPONENT_GLOBALS()

#define COMPONENT_UPP_SELECT_ROOT()		MovieExport
#define COMPONENT_DISPATCH_FILE			"LAMEExportDispatch.h"


#if __APPLE_CC__
//    #include <QuickTime/Components.k.h>	MacOS X 10.1 is missing this from the framework
    #include <CoreServices/Components.k.h>
    #include <QuickTime/QuickTimeComponents.k.h>
    #include <QuickTime/ComponentDispatchHelper.c>
#else
    #include <CoreServices/Components.k.h>
    #include <QuickTimeComponents.k.h>
    #include <ComponentDispatchHelper.c>
#endif

#define	kDefault_BitRateIndex	8
#define	kDefault_sampleRate		44100
#define	kDefault_vbrQuality		5
#define	kDefault_quality		7
#define	vbr_auto				vbr_max_indicator
static unsigned short gBitRates[] = {16, 32, 40, 48, 56, 64, 80, 96, 112, 128, 160, 192, 256, 320};

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//			Component Routines
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
static	void				LAME_GetDefaultSettings(LAME_Settings *settingsP)
{
	settingsP->vbrMode		= vbr_auto;
	settingsP->vbrQuality	= kDefault_vbrQuality;
	settingsP->stereoMode	= NOT_SET;
	settingsP->sampleRate	= kDefault_sampleRate;
	settingsP->bitRate		= gBitRates[kDefault_BitRateIndex];
	settingsP->quality		= kDefault_quality;
}

COMPONENTFUNC LAMEExportOpen(LAMEExportGlobalsPtr globals, ComponentInstance self)
{
    OSErr						result;
    
    globals = (LAMEExportGlobalsPtr)NewPtrClear(sizeof(LAMEExportGlobals));
    if (globals != nil) {	
        // set our storage pointer to our globals
        SetComponentInstanceStorage(self, (Handle) globals);
        globals->self = self;
		LAME_GetDefaultSettings(&globals->settings);
        result = noErr;
    } else {
        result = MemError();
	}
    
    return result;
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
COMPONENTFUNC LAMEExportClose(LAMEExportGlobalsPtr globals, ComponentInstance self)
{
    ComponentResult		result = noErr;
    
    if (globals->tagMappings != NULL)
        DisposeHandle(globals->tagMappings);
    
    if (globals != nil)											// we have some globals
    {
        DisposePtr((Ptr)globals);								// torch them
    }
    
    return result;
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
COMPONENTFUNC LAMEExportVersion(LAMEExportGlobalsPtr globals)
{
	return kLAMECodecHexVersion;
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
COMPONENTFUNC LAMEExportToFile(LAMEExportGlobalsPtr globals, const FSSpec *theFile, 
                               Movie m, Track onlyThisTrack, TimeValue startTime, TimeValue duration)
{
    ComponentResult err;
    AliasHandle alias;
    err = QTNewAlias (theFile, &alias, true);
    if (err == noErr)
    {
        err = MovieExportToDataRef(globals->self, (Handle)alias, rAliasType, m, onlyThisTrack, startTime, duration);
        DisposeHandle((Handle)alias);
    }
    
    return err;
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
COMPONENTFUNC LAMEExportGetSourceMediaType(LAMEExportGlobalsPtr globals, OSType * mediaType)
{
    *mediaType = SoundMediaType;
    
    return noErr;
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
COMPONENTFUNC LAMEExportValidate(LAMEExportGlobalsPtr globals, Movie theMovie, Track onlyThisTrack,
                                 Boolean *valid)
{
    *valid = false;
    
    if (onlyThisTrack != nil)
    {
        OSType	mediaType;
        
        Media media = GetTrackMedia(onlyThisTrack);
        GetMediaHandlerDescription(media, &mediaType, nil, nil);
        
        *valid = (mediaType == SoundMediaType);
    } // check the one track
    else
    {
        long trackCount = GetMovieTrackCount(theMovie);
        int		i;
        
        for (i = 1; i <= trackCount; i++)
        {
            OSType	mediaType;
            Track theTrack = GetMovieIndTrack(theMovie, i);
            Media media = GetTrackMedia(theTrack);
            
            GetMediaHandlerDescription(media, &mediaType, nil, nil);
            
            if (mediaType == SoundMediaType)
            {
                *valid = true;
                break;
            }
        }
    } // check all the tracks
    
    return noErr;
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
COMPONENTFUNC LAMEExportGetFileNameExtension(LAMEExportGlobalsPtr globals, OSType *extension)
{
    *extension = kCodecFormat;
    return noErr;
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
COMPONENTFUNC LAMEExportGetShortFileTypeString(LAMEExportGlobalsPtr globals, Str255 typeString)
{
    typeString[0] = '\3';
    typeString[1] = 'm';
    typeString[2] = 'p';
    typeString[3] = '3';
    
    return noErr;
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
COMPONENTFUNC LAMEExportSetProgressProc(LAMEExportGlobalsPtr globals, MovieProgressUPP proc, long refcon)
{
    globals->progressProc = proc;
    globals->progressRefcon = refcon;
    
    return noErr;
}


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
COMPONENTFUNC LAMEExportAddDataSource(LAMEExportGlobalsPtr globals, OSType trackType, TimeScale scale,
                                      long *trackID, MovieExportGetPropertyUPP getPropertyProc, MovieExportGetDataUPP getDataProc, void *refcon)
{
    ComponentResult err = noErr;
    
    if (globals->numDataSources == 0 && trackType == SoundMediaType)
    {
        globals->source.trackType = trackType;
        globals->source.trackID = 1;
        globals->source.scale = scale;
        globals->source.getPropertyProc = getPropertyProc;
        globals->source.getDataProc = getDataProc;
        globals->source.refcon = refcon;
        
        *trackID = globals->source.trackID;
        
        globals->numDataSources++;
    }
    else
        err = paramErr;
    
    return err;
}

static int mapLAMEError(int vErr)
{
	return paramErr;			//@@@@ do we need to do better than this?
}


static OSErr GetSourceData(LAMEExportGlobalsPtr globals, int numSamples)
{
    globals->getData.recordSize         = sizeof(globals->getData);
    globals->getData.trackID            = globals->source.trackID;
    globals->getData.sourceTimeScale    = globals->source.scale;
    globals->getData.requestedTime      = globals->startTime;
    
    globals->getData.requestedSampleCount = numSamples;
    globals->getData.actualSampleCount  = 0;
    
    return InvokeMovieExportGetDataUPP(globals->source.refcon, &globals->getData, globals->source.getDataProc);
}

static pascal Boolean FillBufferDataProc(SoundComponentDataPtr* data, LAMEExportGlobalsPtr globals)
{
    OSErr	err = noErr;
    Boolean haveMore = false;
    
    if (globals->startTime < globals->endTime && globals->sourceError == noErr)
    {
        if (!globals->didFirstFetch)
        {
            err = GetSourceData(globals, kNumSamplesToFetch);
            if (err != noErr)
                globals->sourceError = err;
        }
        
        globals->didFirstFetch = false;
        
        /* advance the start time */
        globals->startTime = globals->getData.actualTime + (globals->getData.actualSampleCount * globals->getData.durationPerSample);
        
        globals->sourceData.desc.buffer = (Byte *)globals->getData.dataPtr;
        globals->sourceData.desc.sampleCount = globals->getData.actualSampleCount;
        globals->sourceData.bufferSize = globals->getData.dataSize;
        
        haveMore = (err == noErr) && (globals->getData.actualSampleCount != 0);
    }
    else
    {
        globals->sourceData.desc.buffer = nil;
        globals->sourceData.desc.sampleCount = 0;
        globals->sourceData.bufferSize = 0;
    }
    
    *data = (SoundComponentDataPtr)&globals->sourceData;
    return haveMore;
}

static OSErr OpenSoundConverter(LAMEExportGlobalsPtr globals)
{
    OSErr				err;
    SoundDescriptionPtr	sd = *(SoundDescriptionHandle)globals->getData.desc;
    
    // setup input/output format for sound converter
    globals->theInputFormat.flags = kHighQuality | kNonRealTime;		// do a good job on export
    globals->theInputFormat.format = sd->dataFormat;
    globals->theInputFormat.numChannels = sd->numChannels;
    globals->theInputFormat.sampleSize = sd->sampleSize;
    globals->theInputFormat.sampleRate = sd->sampleRate;
    globals->theInputFormat.sampleCount = 0;
    globals->theInputFormat.buffer = NULL;
    globals->theInputFormat.reserved = 0;
    
    globals->theOutputFormat.flags = kHighQuality | kNonRealTime;		// do a good job on export 
    globals->theOutputFormat.format = k16BitNativeEndianFormat;
    
    globals->theOutputFormat.numChannels = 2;   //JS: sd->numChannels;
    globals->theOutputFormat.sampleSize = 16;							// only know how to parse 16 bit sound
    globals->theOutputFormat.sampleRate = sd->sampleRate;
    globals->theOutputFormat.sampleCount = 0;
    globals->theOutputFormat.buffer = NULL;
    globals->theOutputFormat.reserved = 0;
    
    globals->sourceData.desc = globals->theInputFormat;
    globals->sourceData.desc.flags |= kExtendedSoundData;
    globals->sourceData.recordSize = sizeof(ExtendedSoundComponentData);
    globals->sourceData.extendedFlags = kExtendedSoundSampleCountNotValid | kExtendedSoundBufferSizeValid;
    globals->sourceData.bufferSize = 0;
    
    globals->convertOutBuffer = NewPtr(kConvertBufferSize);
    if (globals->convertOutBuffer != nil)
    {
        globals->fillBufferDataUPP = NewSoundConverterFillBufferDataUPP((SoundConverterFillBufferDataProcPtr)FillBufferDataProc);
        if (globals->fillBufferDataUPP)
        {
            err = SoundConverterOpen(&globals->theInputFormat, &globals->theOutputFormat, &globals->mySoundConverter);
            if (err == noErr)
            {
                Handle	extension = NewHandle(0);
                
                // get the "magic" decompression atom
                // This extension to the SoundDescription information stores data specific to a given audio decompressor.
                // Some audio decompression algorithms require a set of out-of-stream values to configure the decompressor
                // which are stored in a siDecompressionParams atom. The contents of the siDecompressionParams atom are dependent
                // on the audio decompressor.
                err = GetSoundDescriptionExtension((SoundDescriptionHandle)globals->getData.desc, &extension, siDecompressionParams);
                if (err == noErr)
                {
                    // set up the sound converters decompresson 'environment' by passing in the 'magic' decompression atom
                    HLock(extension);
                    err = SoundConverterSetInfo(globals->mySoundConverter, siDecompressionParams, *extension);
                    if (siUnknownInfoType == err) {
                        // clear this error, the decompressor didn't
                        // need the decompression atom and that's OK
                        err = noErr;
                    }
                    DisposeHandle(extension);
                }
                else if (err == codecExtensionNotFoundErr)  // hide the failure of GetSoundDescriptionExtension
                    err = noErr;
                
                if (err == noErr)
                    err = SoundConverterBeginConversion(globals->mySoundConverter);
                globals->state = kReadingData;
            }
        }
        else
            err = memFullErr;
    }
    else
        err = memFullErr;
    
    if (err != noErr && globals->convertOutBuffer != nil)
    {
        DisposePtr(globals->convertOutBuffer);
        globals->convertOutBuffer = nil;
    }
    
    if (err != noErr && globals->fillBufferDataUPP)
    {
        DisposeSoundConverterFillBufferDataUPP(globals->fillBufferDataUPP);
        globals->fillBufferDataUPP = nil;
    }
    
    return err;
}

static OSErr CloseSoundConverter(LAMEExportGlobalsPtr globals)
{
    if (globals->fillBufferDataUPP)
    {
        DisposeSoundConverterFillBufferDataUPP(globals->fillBufferDataUPP);
        globals->fillBufferDataUPP = nil;
    }
    
    if (globals->convertOutBuffer)
    {
        DisposePtr(globals->convertOutBuffer);
        globals->convertOutBuffer = nil;
    }
    
    if (globals->mySoundConverter)
    {
        SoundConverterClose(globals->mySoundConverter);
        globals->mySoundConverter = nil;
    }
    
    return noErr;
}


static ComponentResult GetWriteDataHandler(LAMEExportGlobalsPtr globals, Handle dataRef, OSType dataRefType)
{
    ComponentResult	err = cantOpenHandler;
    Component		dh;
    
    globals->dataOffset = 0L;
    dh = GetDataHandler(dataRef, dataRefType, kDataHCanWrite);
    if (dh != nil)
    {
        err = OpenAComponent(dh, &globals->dataHandler);
        if (err == noErr)
        {
            err = DataHSetDataRef(globals->dataHandler, dataRef);
            if (err == noErr)
            {
                err = DataHOpenForWrite(globals->dataHandler);
                if (err == noErr)
                    return noErr;
            }
            CloseComponent(globals->dataHandler);
            globals->dataHandler = nil;
        }
    }
    
    return err;		
}


static ComponentResult WriteOut(LAMEExportGlobalsPtr globals, Ptr p, long len)
{
    ComponentResult err = DataHWrite(globals->dataHandler, p, globals->dataOffset, len, nil, 0);
    globals->dataOffset += len;
    
    return err;
}


static ComponentResult DoHeader(LAMEExportGlobalsPtr globals)
{
    ComponentResult err = noErr;
#ifdef NEVER_DEFINED
/*
 * ogg vorbis is fancier than mp3...perhaps this header isn't necessary either?
 */
    int	vErr;
    ogg_packet header;
    ogg_packet header_comm;
    ogg_packet header_code;
    
    vErr = vorbis_analysis_headerout(&globals->vd,&globals->vc,&header,&header_comm,&header_code);
    if (vErr < 0)
    	return mapLAMEError(vErr);
    
    vErr = ogg_stream_packetin(&globals->os,&header); /* automatically placed in its own page */
    if (vErr < 0)
    	return mapLAMEError(vErr);
    vErr = ogg_stream_packetin(&globals->os,&header_comm);
    if (vErr < 0)
    	return mapLAMEError(vErr);
    vErr = ogg_stream_packetin(&globals->os,&header_code);
    if (vErr < 0)
    	return mapLAMEError(vErr);
    
    /* We don't have to write out here, but doing so makes streaming 
        * much easier, so we do, flushing ALL pages. This ensures the actual
        * audio data will start on a new page
        */
    while(true){
        ogg_page         og; /* one Ogg bitstream page.  Vorbis packets are inside */
        
        int result=ogg_stream_flush(&globals->os, &og);
        if(result==0)
            break;
        
        err = WriteOut(globals, (Ptr)og.header, og.header_len);
        if (err != noErr)
            break;
        err = WriteOut(globals, (Ptr)og.body, og.body_len);
        if (err != noErr)
            break;
    }
#endif
    return err;
}

static ComponentResult ReadConvertedData(LAMEExportGlobalsPtr globals, Ptr *buffer, UInt32 *outputFrames)
{
    ComponentResult	err = noErr;
    UInt32	actualOutputBytes;
    UInt32	outputFlags;
    
    *buffer = globals->convertOutBuffer;
    if (globals->state == kReadingData)
    {
        err = SoundConverterFillBuffer(globals->mySoundConverter,	// a sound converter
                                       globals->fillBufferDataUPP,	// the callback UPP
                                       globals,						// refCon passed to FillDataProc
                                       globals->convertOutBuffer,	// the decompressed data 'play' buffer
                                       kConvertBufferSize,			// size of the 'play' buffer
                                       &actualOutputBytes,		// number of output bytes
                                       outputFrames,			// number of output frames
                                       &outputFlags);			// fillbuffer returned advisor flags
        
        if((outputFlags & kSoundConverterHasLeftOverData) == false) {
            globals->state = kSawEndofData;
            if (*outputFrames == 0)
            {
                goto ConversionDone;
            }
        }
    }
    else
    {
        if (globals->state == kEndedConversion)
            *outputFrames = 0;
        else
        {
ConversionDone:
            globals->state = kEndedConversion;
            err = SoundConverterEndConversion(globals->mySoundConverter, globals->convertOutBuffer, outputFrames, &actualOutputBytes);
            if (actualOutputBytes > (unsigned)kConvertBufferSize)
                DebugStr("\pOoops - overran output buffer");
        }
    }
    
    if (err == noErr && globals->progressProc != nil && globals->sourceMovie != nil)
    {
        TimeValue	progress = globals->startTime - globals->progressStart;
        progress >>= globals->progressShift;
        
        err = InvokeMovieProgressUPP(globals->sourceMovie, movieProgressUpdatePercent, progressOpExportMovie,
                                     FixRatio((short)progress, (short)globals->progressDuration), globals->progressRefcon, globals->progressProc);
    }
    
    return err;
}

static ComponentResult DoBody(LAMEExportGlobalsPtr globals, int numChannels)
{
    int		eos = false;
    ComponentResult	err = paramErr, tempErr;
    Ptr		dataBuffer;
    UInt32	outputFrames;
    int		vErr = noErr;
    
    while(!eos) {
        err = ReadConvertedData(globals, &dataBuffer, &outputFrames);
        
        if (err != noErr || outputFrames == 0) {
            /* end of file.  this can be done implicitly in the mainline,
            but it's easier to see here in non-clever fashion.
            Tell the library we're at end of stream so that it can handle
            the last frame and mark end of stream in the output properly */
            //vorbis_analysis_wrote(&globals->vd,0);
            
            /*
             * Note that docs indicate other calls with the size of the output buffer
             * can take a zero and be smart...I'm hoping that's the case here...if not
             * we'll probably find out when we get a bunch of zero length MP3 files
             */
            vErr = lame_encode_flush(globals->lameflags, (unsigned char *)globals->convertOutBuffer, 0);
            if (vErr < 0)
                return mapLAMEError(vErr);
            
            eos = 1;
            
        }else{
            /* data to encode */
            short *s = (short *)dataBuffer;
#ifdef NEVER_DEFINED
            UInt32 i;
            int	j;
            float **buffer;
            
            /* expose the buffer to submit data */
            buffer=vorbis_analysis_buffer(&globals->vd,outputFrames);
            
            /* uninterleave samples */
            /* because we use the converter to get Native Endian 16-bit samples, we can read efficiently */
            for(i=0; i<outputFrames; i++) {
                for (j = 0; j < numChannels; j++) {
                    buffer[j][i] = ((int)*s) / 32768.f;
                    s++;
                }
            }
	    
            /* tell the library how much we actually submitted */
            vErr = vorbis_analysis_wrote(&globals->vd,i);
#endif
            //vErr = lame_encode_buffer(globals->lameflags, buffer[0], buffer[1], i, globals->convertOutBuffer, 0);
            vErr = lame_encode_buffer_interleaved(globals->lameflags, s, outputFrames, (unsigned char *)globals->convertOutBuffer, 0);
            if (vErr < 0)
                return mapLAMEError(vErr);
        }
        
        /* now jet it into the data handler */
        if (err == noErr || err == userCanceledErr) {
            tempErr = WriteOut(globals, globals->convertOutBuffer, vErr);
            if (tempErr != noErr) {
                return tempErr;
            }
        }
        
#ifdef NEVER_DEFINED
        /* vorbis does some data preanalysis, then divvies up blocks for
            more involved (potentially parallel) processing.  Get a single
            block for encoding now */
        if (err == noErr || err == userCanceledErr)
        {
            while (vorbis_analysis_blockout(&globals->vd, &globals->vb) == 1)
            {
                ogg_packet       op; /* one raw packet of data for decode */
                
                vErr = vorbis_analysis(&globals->vb,NULL);
                if (vErr < 0)
                    return mapLAMEError(vErr);
                
                vErr = vorbis_bitrate_addblock(&globals->vb);
                if (vErr < 0)
                    return mapLAMEError(vErr);
                
                while (vorbis_bitrate_flushpacket(&globals->vd, &op))
                {
                    /* weld the packet into the bitstream */
                    vErr = ogg_stream_packetin(&globals->os, &op);
                    if (vErr < 0)
                        return mapLAMEError(vErr);
                    /* write out pages (if any) */
                    while(!eos)
                    {
                    	ogg_page         og; /* one Ogg bitstream page.  Vorbis packets are inside */
                        
                        int result = ogg_stream_pageout(&globals->os, &og);
                        if(result == 0)
                            break;
                        tempErr = WriteOut(globals, (Ptr)og.header, og.header_len);
                        if (tempErr != noErr)
                            return tempErr;
                        tempErr = WriteOut(globals, (Ptr)og.body, og.body_len);
                        if (tempErr != noErr)
                            return tempErr;
                        
                        /* this could be set above, but for illustrative purposes, I do
                            it here (to show that vorbis does know where the stream ends) */
                        
                        if(ogg_page_eos(&og))
                            eos=1;
                    }
                }
            }
        }
#endif
    }
    
    return err;
}

#ifdef NEVER_DEFINED
static char *GetLAMETagMapping(LAMEExportGlobalsPtr globals, OSType udType)
{
    typedef struct {
        OSType                osType;
        char                  tag[1];
    } LAMETagMappingEntry;
    
    typedef struct {
        long                count;
        LAMETagMappingEntry table;
    } LAMETagMapping;
    
    LAMETagMapping      *tagTable;
    LAMETagMappingEntry *p;
    long                count;
    OSType              endianUdType = EndianU32_NtoB(udType);
    
    if (globals->tagMappings == NULL)
    {
        OSErr err = GetComponentResource((Component)globals->self, 'TagM', kDecompressorResID, &globals->tagMappings);
        if (err != noErr)
            return NULL;
        HLock(globals->tagMappings);
    }
    
    tagTable = *(LAMETagMapping **)globals->tagMappings;
    count = EndianU32_BtoN(tagTable->count);
    p = &tagTable->table;
    
    while (count--)
    {
        if (p->osType == endianUdType)
            return p->tag;
        p = (LAMETagMappingEntry *)((char *)p + ((sizeof(LAMETagMappingEntry) + strlen(p->tag)) & 0xfffffffc));
    }
    
    return NULL;	
}
#endif


#if PRAGMA_STRUCT_ALIGN
    #pragma options align=mac68k
#elif PRAGMA_STRUCT_PACKPUSH
    #pragma pack(push, 2)
#elif PRAGMA_STRUCT_PACK
    #pragma pack(2)
#endif

typedef struct UserItemRecord {
	short	length;
	short	regionCode;
	char 	data[1];
} UserItemRecord, * UserItemRecordPtr;

#if PRAGMA_STRUCT_ALIGN
    #pragma options align=reset
#elif PRAGMA_STRUCT_PACKPUSH
    #pragma pack(pop)
#elif PRAGMA_STRUCT_PACK
    #pragma pack()
#endif


#if TARGET_OS_WIN32
static int region2codepage[] = {
	/*   0 verUS */              10000,
	/*   1 verFrance */          10000,
	/*   2 verBritian */         10000,
	/*   3 verGermany */         10000,
	/*   4 verItaly */           10000,
	/*   5 verNetherlands */     10000,
	/*   6 verFlemish */         10000,
	/*   7 verSweden */          10000,
	/*   8 verSpain */           10000,
	/*   9 verDenmark */         10000,
	/*  10 verPortugal */        10000,
	/*  11 verFrCanada */        10000,
	/*  12 verNorway */          10000,
	/*  13 verIsrael */          10005,
	/*  14 verJapan */           10001,
	/*  15 verAustralia */       10000,
	/*  16 verArabic */          10004,
	/*  17 verFinland */         10000,
	/*  18 verFrSwiss */         10000,
	/*  19 verGrSwiss */         10000,
	/*  20 verGreece */          10006,
	/*  21 verIceland */         10079,
	/*  22 verMalta */           0,        /* unknown */
	/*  23 verCyprus */          0,
	/*  24 verTurkey */          10081,
	/*  25 verYugoCroation */    10082,
	/*  26 verNetherlandsComma */ -1,   /* not used */
	/*  27 verBelgiumLuxPoint */  -1,
	/*  28 verCanadaComma */      -1,
	/*  29 verCanadaPoint */      -1,
	/*  30 vervariantPortugal */  -1,
	/*  31 vervariantNorway */    -1,
	/*  32 vervariantDenmark */   -1,
	/*  33 verIndiaHindi */      0,
	/*  34 verPakistanUrdu */    0,
	/*  35 verTurkishModified */ 10081,
	/*  36 verItalianSwiss */    10000,
	/*  37 verInternational */   10000,
	/*  38 not used */           -1,
	/*  39 verRomania */         10010,
	/*  40 verGreecePoly */      10006,
	/*  41 verLithuania */       10029,
	/*  42 verPoland */          10029,
	/*  43 verHungary */         10029,
	/*  44 verEstonia */         10029,
	/*  45 verLatvia */          10029,
	/*  46 verSami */            -1,
	/*  47 verFaroeIsl */        10079,
	/*  48 verIran */            -1,
	/*  49 verRussia */          10007,
	/*  50 verIreland */         10000,
	/*  51 verKorea */           10003,
	/*  52 verChina */           10008,
	/*  53 verTaiwan */          10002,
	/*  54 verThailand */        10021,
	/*  55 verScriptGeneric */   -1,
	/*  56 verCzech */           10029,
	/*  57 verSlovak */          10029,
	/*  58 verFarEastGeneric */  -1,
	/*  59 verMagyar */          -1,
	/*  60 verBengali */         -1,
	/*  61 verByeloRussian */    10007,
	/*  62 verUkraine */         10017,
	/*  63 not used */           -1,
	/*  64 not used */           -1,
	/*  65 verSerbian */         10029,
	/*  66 verSlovenian */       10029,
	/*  67 verMacedonian */      10029,
	/*  68 verCroatia */         10082,
	/*  69 not used */           -1,
	/*  70 verGermanReformed */  10000,
	/*  71 verBrazil */          10000,
	/*  72 verBulgaria */        10007,
	/*  73 verCatalonia */       10000,
	/*  74 verMultiLingual */    -1,
	/*  75 verScottishGaelic */  10000,
	/*  76 ManxGaelic */         10000,
	/*  77 verBrenton */         0,
	/*  78 verNunavut */         0,
	/*  79 verWelsh */           10000,
	/*  80 not used */           -1,
	/*  81 verIrishGaelicScript */10000,
	/*  82 verEngCanada */       10000,
	/*  83 verBhutan */          -1,
	/*  84 verArmenian */        10007,
	/*  85 verGeorgian */        10007,
    /*  86 verSpLatinAmerica */  10000,
	/*  87 not used */           -1,
	/*  88 verTonga */           10000,
	/*  89 not used */           -1,
	/*  90 not used */           -1,
	/*  91 verFrenchUniversal */ 10000,
	/*  92 verAustria */         10000,
	/*  93 not used */           -1,
	/*  94 verGujarati */        0,
	/*  95 verPunjabi */         0,
	/*  96 verIndiaUrdu */       0,
	/*  97 verVietnam */         1258,
	/*  98 verFrBelgium */       10000,
	/*  99 verUzbek */           0,
    /* 100 verSingapore */       0,
    /* 101 verNynorsk */         0,
    /* 102 verAfrikaans */       10000,
    /* 103 verEsperanto */       10000,
    /* 104 verMarathi */         0,
    /* 105 verTibetan */         0,
    /* 106 verNepal */           0,
    /* 107 verGreenland */       10000
};
#endif

typedef struct RegionToUnicodeRecord 
{
#if TARGET_OS_MAC
	short			curRegionCode;
	TextEncoding	sourceEncoding;
	TextEncoding	destEncoding;
	TECObjectRef	converter;
#else
	Handle			tempMem;
#endif
} RegionToUnicodeRecord, * RegionToUnicodeRecordPtr;

void InitRegionToUnicodeRecord(RegionToUnicodeRecordPtr rtn)
{
#if TARGET_OS_MAC
	rtn->curRegionCode = -1;
	rtn->sourceEncoding = 0;
	rtn->destEncoding = CreateTextEncoding(kTextEncodingUnicodeV3_0, kUnicodeNoSubset, kUnicodeUTF8Format);
	rtn->converter = NULL;
#else
	rtn->tempMem = NULL;
#endif
}

OSStatus ConvertToUnicode(RegionToUnicodeRecordPtr rtn, Ptr text, unsigned int length, RegionCode regionCode, Handle *h)
{
    OSStatus err = noErr;
#if TARGET_OS_MAC
    Handle temp;
    long tempLen = length * 5;
    ByteCount actualIn, actualOut, flushOut;
    
    *h = NULL;
    
    if (tempLen < 32)
        tempLen = 32;
    
    if (regionCode != rtn->curRegionCode)
    {
        if (rtn->converter != NULL)
        {
            TECDisposeConverter(rtn->converter);
            rtn->converter = NULL;
        }
    }
    
    if (rtn->converter == NULL)
    {
        err = UpgradeScriptInfoToTextEncoding(kTextScriptDontCare, kTextLanguageDontCare,
                                              regionCode, NULL, &rtn->sourceEncoding);
        
        if (err == noErr)
        {
            err = TECCreateConverter(&rtn->converter, rtn->sourceEncoding, rtn->destEncoding);
            if (err != noErr)
                return err;
        }
        rtn->curRegionCode = regionCode;
    } // make a converter for this encoding
    
    // now we have the right converter, call it
    
    temp = NewHandle(tempLen);
    if (temp == NULL)
        return memFullErr;
    
    while (1)
    {
        HLock(temp);
        
        actualIn = 0;
        err = TECConvertText(rtn->converter, (ConstTextPtr)text, length, &actualIn, (TextPtr)*temp, tempLen, &actualOut);
        if (length == actualIn)
        {
            err = TECFlushText(rtn->converter, (TextPtr) ((*temp) + actualOut), tempLen - actualOut, &flushOut);
            HUnlock(temp);
            if (err == noErr)
            {
                SetHandleSize(temp, actualOut + flushOut);
                *h = temp;
                temp = NULL;
            }
            break;
        }
        else
        {
            HUnlock(temp);
            tempLen += length;
            SetHandleSize(temp, tempLen);
            err = MemError();
            if (err != noErr)
                break;	
        }
    }
    
    if (temp != NULL)
        DisposeHandle(temp);
#else
    int sizeNeeded = 0;
    
    if (rtn->tempMem == NULL)
        rtn->tempMem = NewHandle(1024);
    
    
    while (sizeNeeded == 0)
    {
        sizeNeeded = MultiByteToWideChar(region2codepage[regionCode], 0, text, length, (unsigned short *)*rtn->tempMem, GetHandleSize(rtn->tempMem) / 2);
        err = GetLastError();
        if (sizeNeeded == 0 && err == ERROR_INSUFFICIENT_BUFFER)
        {
            sizeNeeded = MultiByteToWideChar(region2codepage[regionCode], 0, text, length, NULL, 0);
            SetHandleSize(rtn->tempMem, sizeNeeded * 2);
        }
        else
            break;
    }
    
    if (sizeNeeded > 0)
    {
        int i;
        unsigned char *p = *rtn->tempMem;
        *h = NewHandle(sizeNeeded * 3);
        
        if (*h != NULL)
        {
            unsigned char *out = **h;
            
            /* UTF-8 encode the 16-bit unicode characters */
            for (i = 0; i < sizeNeeded; i++)
            {
                unsigned short w = *p++;
                w |= (*p++ << 8);
                
                if (w < (unsigned short) 0x0080)
                    *out++ = (unsigned char) w;
                else if (w < (unsigned short) 0x0800)
                {
                    *out++ = (unsigned short) 0x00c0 | ((w >> 6) & (unsigned short) 0x001f); // 5 bits
                    *out++ = (unsigned short) 0x0080 | ( w       & (unsigned short) 0x003f); // 6 bits
                }
                else
                {
                    *out++ = (unsigned short) 0x00e0 | ((w >> 12) & (unsigned short) 0x000f); // 4 bits
                    *out++ = (unsigned short) 0x0080 | ((w >>  6) & (unsigned short) 0x003f); // 6 bits
                    *out++ = (unsigned short) 0x0080 | ( w        & (unsigned short) 0x003f); // 6 bits
                }
            }
            
            SetHandleSize(*h, out - **h);
            err = MemError();
            if (err != noErr)
            {
                DisposeHandle(*h);
                *h = NULL;
            }
        }
        else
            err = MemError();
    }
    else
        err = memFullErr;
#endif
    
    return err;
}

void DisposeRegionToUnicodeRecord(RegionToUnicodeRecordPtr rtn)
{
#if TARGET_OS_MAC
    if (rtn->converter != NULL)
    {
        TECDisposeConverter(rtn->converter);
        rtn->converter = NULL;
    }
#else
    if (rtn->tempMem != NULL)
        DisposeHandle(rtn->tempMem);
#endif
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

static ComponentResult AddComments(LAMEExportGlobalsPtr globals, UserData ud)
{
    ComponentResult	err = noErr;
#ifdef NEVER_DEFINED
    OSType			udType = 0;
    RegionToUnicodeRecord rtn;
    
    InitRegionToUnicodeRecord(&rtn);
    
    while (udType = GetNextUserDataType(ud, udType))
    {
        if ((udType & 0xff000000) == FOUR_CHAR_CODE(0xa9000000))
        {
            int	count = CountUserDataType(ud, udType);
            int i = 1;
            
            char *tag = GetLAMETagMapping(globals, udType);
            Handle  tagHandle = NULL;
            
            if (tag == NULL)
                tag = "COMMENT=";
            
            while (i <= count)
            {
                Handle	h = NewHandle(1);
                
                err = GetUserData(ud, h, udType, i);
                
                if (err == noErr)
                {
                    Handle unicode;
                    UserItemRecordPtr uir;
                    
                    HLock(h);
                    uir = (UserItemRecordPtr) *h;
                    err = ConvertToUnicode(&rtn, uir->data, EndianU16_BtoN(uir->length), EndianU16_BtoN(uir->regionCode), &unicode);
                    
                    if (err == noErr)
                    {
                        PtrToHand(tag, &tagHandle, strlen(tag));
                        HandAndHand(unicode, tagHandle);
                        DisposeHandle(unicode);
                        PtrAndHand("\0", tagHandle, 1);		// make sure it's null terminated
                        HLock(tagHandle);
                        
                        vorbis_comment_add(&globals->vc, *tagHandle);
                        DisposeHandle(tagHandle);
                    }
                } // if got the text ok
                DisposeHandle(h);
                i++;
            }
        } // if it's a text item
    }
    
    DisposeRegionToUnicodeRecord(&rtn);
    
#endif 

    return err;
}


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
static	void	SanitizeSettings(
	const LAME_Settings	*inP, 
	LAME_Settings		*outP)
{
	*outP = *inP;
	
	if (outP->sampleRate == 0) {
		outP->sampleRate = kDefault_sampleRate;
	}
	
	if (outP->stereoMode == NOT_SET) {
		outP->stereoMode = JOINT_STEREO;
	}
	
	if (outP->vbrMode == vbr_auto) {
		outP->vbrMode = vbr_off;
	}

	if (outP->bitRate == 0) {
		outP->bitRate = gBitRates[kDefault_BitRateIndex];
	}
}

inline static void	LogSettings(
	const LAME_Settings	*inP, 
	bool				gettingB, 
	OSStatus			err)
{
	#if 0
	printf(
		"%s LAME Settings: %s (%ld)\n", 
		gettingB ? "Getting" : "Setting", 
		err ? "ERROR:" : "No error", 
		(long)err);

	LAME_Settings	outSettings;

	SanitizeSettings(inP, &outSettings);

	printf("Set vbrMode    : %ld\n", (long)outSettings.vbrMode);
	printf("Set vbrQuality : %ld\n", (long)outSettings.vbrQuality);
	printf("Set stereoMode : %ld\n", (long)outSettings.stereoMode);
	printf("Set sampleRate : %ld\n", (long)outSettings.sampleRate);
	printf("Set bitRate    : %ld\n", (long)outSettings.bitRate);
	printf("Set quality    : %ld\n", (long)outSettings.quality);
	#endif
}

COMPONENTFUNC LAMEExportFromProceduresToDataRef(LAMEExportGlobalsPtr globals, Handle dataRef, OSType dataRefType)
{
    int   vErr;
    ComponentResult err = noErr;
  		
    if (globals->progressProc && globals->sourceMovie != nil)
    {
        err = InvokeMovieProgressUPP(globals->sourceMovie, movieProgressOpen, progressOpExportMovie,
                                     0, globals->progressRefcon, globals->progressProc);
    }
    
    if (err == 0)
    {
        globals->sourceError = noErr;
        err = GetSourceData(globals, kNumSamplesToFetch);
        if (err == 0)
        {
            globals->didFirstFetch = true;		// mark the first fetch as being done
            
            if (globals->getData.descType == SoundMediaType)
            {
                // next get the write data handler
                err = GetWriteDataHandler(globals, dataRef, dataRefType);
                if (err == noErr)
                {
                    err = OpenSoundConverter(globals);
                    if (err == noErr)
                    {
                        /*
                         * This has to do with the options they selected...we'll handle that later
                        Boolean useManaged = globals->settings.minBitRate != -1 ||
                        globals->settings.desiredBitRate != -1 ||
                        globals->settings.maxBitRate != -1;
                         */
                        
                        SoundDescriptionPtr	sd = *(SoundDescriptionHandle)globals->getData.desc;
                        
                        globals->lameflags = lame_init();
						
						LAME_Settings	outSettings;
						SanitizeSettings(&globals->settings, &outSettings);

                        lame_set_in_samplerate(globals->lameflags, (sd->sampleRate >> 16));
						lame_set_out_samplerate(globals->lameflags, outSettings.sampleRate);
						lame_set_mode(globals->lameflags, outSettings.stereoMode);
						lame_set_VBR(globals->lameflags, outSettings.vbrMode);
						lame_set_VBR_q(globals->lameflags, outSettings.vbrQuality);
						lame_set_brate(globals->lameflags, outSettings.bitRate);
						lame_set_num_channels(globals->lameflags, globals->settings.stereoMode == MONO ? 1 : 2);
						lame_set_quality(globals->lameflags, outSettings.quality);
						
                        vErr = lame_init_params(globals->lameflags);
                        
                        /* fnfErr
                         * init and set some settings perhaps?
                        vorbis_info_init(&globals->vi);
                        
                        if (useManaged)
                            vErr = vorbis_encode_init(&globals->vi, sd->numChannels, (sd->sampleRate >> 16), globals->settings.maxBitRate,
                                                      globals->settings.desiredBitRate, globals->settings.minBitRate);
                        else
                            vErr = vorbis_encode_init_vbr(&globals->vi, sd->numChannels, (sd->sampleRate >> 16), (float)(globals->settings.quality / 100.0));
                        */
                        if (vErr == 0)
                        {
                            /*
                             * looks like stuff to handle comments and what would typically go into the ID3 tags
                            vorbis_comment_init(&globals->vc);
                            */
                            
                            if (globals->sourceMovie)
                            {
                                UserData ud = GetMovieUserData(globals->sourceMovie);
                                if (ud)
                                {
                                    AddComments(globals, ud);
                                }
                            }
                            // Vorbis comments are UTF-8
                            //vorbis_comment_add(&globals->vc,"SOFTWARE=QuickTime Exporter " kLAMECodecVersion);
                            if (vErr == 0)
                            {
                                /*
                                 * I think this stuff just inits some more things...perhaps we don't have to worry
                                 * about this either?
                                // set up the analysis state and auxiliary encoding storage //
                                vorbis_analysis_init(&globals->vd,&globals->vi);
                                vorbis_block_init(&globals->vd,&globals->vb);
                                
                                // set up our packet->stream encoder //
                                ogg_stream_init(&globals->os, TickCount());		// @@@@ pick a better random serial # //
                                */
                                
                                err = DoHeader(globals);
                                if (err == noErr)
                                {
                                    err = DoBody(globals, sd->numChannels);
                                } // if header worked
                                
                                /*
                                 * closing things down
                                vorbis_comment_clear(&globals->vc);		// don't need comments after the header is written out //
                                ogg_stream_clear(&globals->os);
                                vorbis_block_clear(&globals->vb);
                                vorbis_dsp_clear(&globals->vd);
                                */
                            }
                        } else {
							printf("error with init params\n");
                            err = mapLAMEError(vErr);
						}
                        
                        lame_close(globals->lameflags);
                        /*
                         * shut it down
                        vorbis_info_clear(&globals->vi);
                         */
                        
                        CloseSoundConverter(globals);
                    } // open sound converter
                    
                    CloseComponent(globals->dataHandler);
                } // open dataHandler
            }
            else
                err = paramErr;		// can only handle 16-bit big endian right now....
        }
    }
    
    if (globals->progressProc && globals->sourceMovie != nil)
    {
        InvokeMovieProgressUPP(globals->sourceMovie, movieProgressClose, progressOpExportMovie,
                               0, globals->progressRefcon, globals->progressProc);
    }
    
    return err;
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
COMPONENTFUNC LAMEExportToDataRef(LAMEExportGlobalsPtr globals, Handle dataRef, OSType dataRefType, 
                                  Movie theMovie, Track onlyThisTrack, TimeValue startTime, TimeValue duration)
{
    ComponentResult 			err;
    ComponentDescription 		cd;
    ComponentInstance 			ci;
    ComponentRecord 			*movieExporter;
    TimeScale 					scale;
    MovieExportGetPropertyUPP 	getPropertyProc = nil;
    MovieExportGetDataUPP 		getDataProc = nil;
    void 						*refCon;
    long 						trackID;
    
    cd.componentType = MovieExportType;
    cd.componentSubType = 'MooV';
    cd.componentManufacturer = 0;
    cd.componentFlags = canMovieExportFromProcedures;
    cd.componentFlagsMask = canMovieExportFromProcedures;
    
    movieExporter = FindNextComponent(nil, &cd);
    
    if (movieExporter != nil)
    {
        err = OpenAComponent(movieExporter, &ci);
        if (err == noErr)
        {
            err = MovieExportNewGetDataAndPropertiesProcs(ci, SoundMediaType, &scale, theMovie, onlyThisTrack,
                                                          startTime, duration, &getPropertyProc, &getDataProc, &refCon);
            if (err == noErr)
            {
                err = MovieExportAddDataSource(globals->self, SoundMediaType, scale, &trackID, getPropertyProc, getDataProc, refCon);
                if (err == noErr)
                {
                    TimeRecord	tr;
                    
                    globals->sourceMovie = theMovie;
                    
                    /* convert startTime and endTime to the time scale that will be used for the sound
                        we get.
                        */
                    tr.value.hi = 0;
                    tr.value.lo = startTime;
                    tr.scale = GetMovieTimeScale(theMovie);
                    tr.base = nil;
                    ConvertTimeScale(&tr, scale);
                    globals->startTime   = tr.value.lo;
                    
                    tr.value.hi = 0;
                    tr.value.lo = startTime + duration;
                    tr.scale = GetMovieTimeScale(theMovie);
                    ConvertTimeScale(&tr, scale);
                    globals->endTime 	 = tr.value.lo;
                    
                    globals->progressStart = globals->startTime;
                    globals->progressDuration = globals->endTime - globals->startTime;
                    
                    while (globals->progressDuration > 32767)
                    {
                        globals->progressShift++;
                        globals->progressDuration >>= 1;
                    }
                    
                    if (err == noErr)
                        err = MovieExportFromProceduresToDataRef(globals->self, dataRef, dataRefType);
                }
                
                MovieExportDisposeGetDataAndPropertiesProcs(ci, getPropertyProc, getDataProc, refCon);
            }
            
            CloseComponent(ci);
        }
    }
    else
        err = cantOpenHandler;		// can't open the default movie exporter
    
    return err;
}

#if TARGET_OS_MAC
//include "OutputSettingsDialog.c"
#elif TARGET_OS_WIN32
#include "WinOutputSettings.cpp"
#else
#error "unsupported platform"
#endif

// SettingsWindowEventHandler
//		Carbon Event handler for the Modal Settings Dialog.
static pascal OSStatus SettingsWindowEventHandler(
	EventHandlerCallRef inHandler, EventRef inEvent, void *inUserData)
{
    OSStatus				result = eventNotHandledErr;
    WindowRef				window = ActiveNonFloatingWindow();
    LAMEExportGlobalsPtr	globals = (LAMEExportGlobalsPtr)(inUserData);
    HICommand				command;
    
    if (NULL == window) goto bail;
    
    GetEventParameter(
		inEvent, kEventParamDirectObject, typeHICommand, 
		NULL, sizeof(HICommand), NULL, &command);
    
    switch (command.commandID)  {

		case kHICommandOK: {
			ControlRef	control_ref_A[kControls_NUMTYPES];
			UInt32		ctrlValue;
			
			//	get all the controls
			short i;
			for (i = kControls_START; i < kControls_NUMTYPES; i++) {
				ControlID		cid = {'PYEh', i};
				
				GetControlByID(window, &cid, &control_ref_A[i]);
			}
			
			//	k_vbrMode
            ctrlValue = GetControl32BitValue(control_ref_A[k_vbrMode]);
            switch (ctrlValue) {
                default:	globals->settings.vbrMode = vbr_auto;		break;
                case 3:		globals->settings.vbrMode = vbr_off;		break;
                case 4:		globals->settings.vbrMode = vbr_abr;		break;
                case 5:		globals->settings.vbrMode = vbr_rh;			break;
            }
			
			//	k_vbrQuality
			globals->settings.vbrQuality = GetControl32BitValue(control_ref_A[k_vbrQuality]) - 1;
			
			// k_stereoMode
            ctrlValue = GetControl32BitValue(control_ref_A[k_stereoMode]);
			switch (ctrlValue) {
				default:	globals->settings.stereoMode = NOT_SET;			break;
				case 3:		globals->settings.stereoMode = MONO;			break;
				case 4:		globals->settings.stereoMode = STEREO;			break;
				case 5:		globals->settings.stereoMode = JOINT_STEREO;	break;
			}
			
			//	k_sampleRate
            ctrlValue = GetControl32BitValue(control_ref_A[k_sampleRate]);
            switch (ctrlValue) {
                case 3:		globals->settings.sampleRate = 8000;		break;				
                case 4:		globals->settings.sampleRate = 11025;		break;
                case 5:		globals->settings.sampleRate = 12000;		break;				
                case 6:		globals->settings.sampleRate = 16000;		break;				
                case 7:		globals->settings.sampleRate = 22050;		break;			
                case 8:		globals->settings.sampleRate = 24000;		break;			
                case 9:		globals->settings.sampleRate = 32000;		break;			
				default:	globals->settings.sampleRate = 44100;		break;
                case 11:	globals->settings.sampleRate = 48000;		break;
            }

			//	k_bitRate
            ctrlValue = GetControl32BitValue(control_ref_A[k_bitRate]) - 1;
			if (ctrlValue >= 0 && ctrlValue < sizeof(gBitRates) / sizeof(gBitRates[0]))
			globals->settings.bitRate = gBitRates[ctrlValue];

			//	k_quality
			globals->settings.quality = GetControl32BitValue(control_ref_A[k_quality]) - 1;

			globals->canceled = false;
			QuitAppModalLoopForWindow(window);
			result = noErr;
		break;
	}
	
	case kHICommandCancel: {
		globals->canceled = true;
		QuitAppModalLoopForWindow(window);
		result = noErr;
		break;
	}
            
	default:
            break; 
    }
    
bail:
	return result;
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
COMPONENTFUNC LAMEExportDoUserDialog(LAMEExportGlobalsPtr globals, Movie theMovie, Track onlyThisTrack,
                                     TimeValue startTime, TimeValue duration, Boolean *canceled)
{
	#pragma unused(theMovie, onlyThisTrack, startTime, duration)
    OSErr				err = resFNotFound;
    CFBundleRef			bundle = NULL;
    IBNibRef			nibRef = NULL;
    WindowRef			window = NULL;
    int					ctrlValue;
    CGrafPtr			savedPort;
    Boolean				portChanged = false;
    EventTypeSpec		eventList[] = {{kEventClassCommand, kEventCommandProcess}};
    EventHandlerUPP		settingsWindowEventHandlerUPP = NewEventHandlerUPP(SettingsWindowEventHandler);
	ControlRef			control_ref_A[kControls_NUMTYPES];
    
    bundle = CFBundleGetBundleWithIdentifier(CFSTR("com.pyehouse.LAMEEncoder"));
    if (NULL == bundle) goto bail;
    
    err = CreateNibReferenceWithCFBundle(bundle, CFSTR("LAMEComponent"), &nibRef);
    if (err) goto bail;
    
    err = CreateWindowFromNib(nibRef, CFSTR("Window"), &window);
    if (err) goto bail;
    
    portChanged = QDSwapPort(GetWindowPort(window), &savedPort);
    
    *canceled = false;
    
	//	get all the controls
	short i;
	for (i = kControls_START; i < kControls_NUMTYPES; i++) {
		ControlID		cid = { 'PYEh', i };
		
		GetControlByID(window, &cid, &control_ref_A[i]);

		/*
			Un-Check the default items as set up by IB - if we don't do this here 
			the pop-up menu ends up having both the current selection as well as 
			the defaults checked at the same time - lame!
		*/
		CheckMenuItem(GetControlPopupMenuHandle(control_ref_A[i]), 1, false);
	}

	//	k_vbrMode
	switch (globals->settings.vbrMode) {
		default:		ctrlValue = 1;	break;
		case vbr_off:	ctrlValue = 3;	break;
		case vbr_abr:	ctrlValue = 4;	break;
		case vbr_rh:	ctrlValue = 5;	break;
	}
	SetControl32BitValue(control_ref_A[k_vbrMode], ctrlValue);

	//	k_vbrQuality
	SetControl32BitValue(control_ref_A[k_vbrQuality], globals->settings.vbrQuality + 1);
	
	// k_stereoMode
	switch (globals->settings.stereoMode) {
		default:			ctrlValue = 1;	break;
		case MONO:			ctrlValue = 3;	break;
		case STEREO:		ctrlValue = 4;	break;
		case JOINT_STEREO:	ctrlValue = 5;	break;
	}
	SetControl32BitValue(control_ref_A[k_stereoMode], ctrlValue);

    // k_sampleRate
    switch (globals->settings.sampleRate) {
		default:	ctrlValue = 1;	break;
		case 8000:	ctrlValue = 3;	break;
		case 11025:	ctrlValue = 4;	break;	
		case 12000:	ctrlValue = 5;	break;
		case 16000:	ctrlValue = 6;	break;		
		case 22050:	ctrlValue = 7;	break;	
		case 24000:	ctrlValue = 8;	break;
		case 32000:	ctrlValue = 9;	break;
		case 44100:	ctrlValue = 10;	break;
		case 48000:	ctrlValue = 11;	break;
    }
    SetControl32BitValue(control_ref_A[k_sampleRate], ctrlValue);
    
	//	k_bitRate
	for (ctrlValue = 0; ctrlValue < (sizeof(gBitRates) / sizeof(gBitRates[0])); ctrlValue++) {
		if (gBitRates[ctrlValue] == globals->settings.bitRate) {
			SetControl32BitValue(control_ref_A[k_bitRate], ctrlValue + 1);
			break;
		}
	}
    
	//	k_quality
	SetControl32BitValue(control_ref_A[k_quality], globals->settings.quality + 1);

    InstallWindowEventHandler(window, settingsWindowEventHandlerUPP, GetEventTypeCount(eventList), eventList, globals, NULL); 
    ShowWindow(window);
    RunAppModalLoopForWindow(window);
    *canceled = globals->canceled;
bail:
    if (window) {
	if (portChanged) {
	    QDSwapPort(savedPort, NULL);
	}
	DisposeWindow(window);
    }
    
    if (settingsWindowEventHandlerUPP) {
        DisposeEventHandlerUPP(settingsWindowEventHandlerUPP);
    }
    
    if (nibRef) {
        DisposeNibReference(nibRef);
    }
    
    return err;
}


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#define		ERR(_x)		if (!err) err = (_x)

typedef	struct {
	UInt32	key;
	long	value;
} AtomType;

COMPONENTFUNC	LAMEExportGetSettingsAsAtomContainer(
	LAMEExportGlobalsPtr	globals, 
	QTAtomContainer			*settings)
{
	ComponentResult		err			= QTNewAtomContainer(settings);
	QTAtom				lameAtom	= kParentAtomIsContainer;
	long				valueL;
	
	// do this when we add settings options add the kVorbisAtomType child
	ERR(QTInsertChild(
		*settings, kParentAtomIsContainer, kLAMESettingsAtom,
		1000, 0, 0, nil, &lameAtom));

    valueL = EndianS32_NtoB(kExporterVersion);
	ERR(QTInsertChild(
		*settings, lameAtom, kLAMESettingsVersionAtom,
		0, 0, sizeof(valueL), &valueL, nil));
	
	/******************************************/
	AtomType	atoms[6];
	
	atoms[0].key	= kLAMESettings_vbrMode;
	atoms[0].value	= globals->settings.vbrMode;

	atoms[1].key	= kLAMESettings_vbrQuality;
	atoms[1].value	= globals->settings.vbrQuality;

	atoms[2].key	= kLAMESettings_stereoMode;
	atoms[2].value	= globals->settings.stereoMode;

	atoms[3].key	= kLAMESettings_sampleRate;
	atoms[3].value	= globals->settings.sampleRate;

	atoms[4].key	= kLAMESettings_bitRate;
	atoms[4].value	= globals->settings.bitRate;

	atoms[5].key	= kLAMESettings_quality;
	atoms[5].value	= globals->settings.quality;
	
	short i;
	for (i = 0; !err && i < 6; i++) {
		long	valueL = EndianS32_NtoB(atoms[i].value);
	
		err = QTInsertChild(
			*settings, lameAtom, atoms[i].key,
			0, 0, sizeof(valueL), &valueL, nil);
	}

	/**********************************************************/
	if (err != noErr) {
		QTDisposeAtomContainer(*settings);
		settings = NULL;
	}

	LogSettings(&globals->settings, true, err);

	return err;
}

static ComponentResult GetSettingsHelper(QTAtomContainer settings, QTAtom lameAtom, OSType type, long *value)
{
    ComponentResult	err = noErr;
    QTAtom	child;
    
    child = QTFindChildByIndex(settings, lameAtom, type, 1, nil);
    if (child)
    {
        long	temp;
        
        err = QTCopyAtomDataToPtr(settings, child, false, sizeof(temp), &temp, nil);
        
        if (err == noErr)
            *value = EndianS32_BtoN(temp);
    }
    
    return err;
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// The 'sean' portion of the atom must be the container atom, since we don't
//

/*
	you are SETting our internal settings FROM those passed in by the Atom
*/
static	ComponentResult		LAME_SetSettingsFromAtom(QTAtomContainer qtAtom, LAME_Settings *settingsP)
{
    ComponentResult	err = noErr;
    QTAtom			lameAtom = QTFindChildByIndex(qtAtom, kParentAtomIsContainer, kLAMESettingsAtom, 1, nil);
    
    if (lameAtom) {
		long		valueL;

        ERR(GetSettingsHelper(qtAtom, lameAtom, kLAMESettings_vbrMode,		&valueL));
		settingsP->vbrMode = valueL;
		
        ERR(GetSettingsHelper(qtAtom, lameAtom, kLAMESettings_vbrQuality,	&settingsP->vbrQuality));

        ERR(GetSettingsHelper(qtAtom, lameAtom, kLAMESettings_stereoMode,	&valueL));
		settingsP->stereoMode = valueL;

        ERR(GetSettingsHelper(qtAtom, lameAtom, kLAMESettings_sampleRate,	&settingsP->sampleRate));
        ERR(GetSettingsHelper(qtAtom, lameAtom, kLAMESettings_bitRate,		&settingsP->bitRate));
        ERR(GetSettingsHelper(qtAtom, lameAtom, kLAMESettings_quality,		&settingsP->quality));
    }
    
    return err;
}

COMPONENTFUNC LAMEExportSetSettingsFromAtomContainer(LAMEExportGlobalsPtr globals, QTAtomContainer settings)
{
    ComponentResult		err = LAME_SetSettingsFromAtom(settings, &globals->settings);

	LogSettings(&globals->settings, false, err);
    return err;
}

